use std::ops::Deref;
use crate::artifacts::Artifact;
use crate::level::{Level, StructureElement};
use crate::level_generator::LevelGenerator;
use crate::monster::Monster;
use crate::player::Player;
use crate::position::Position;

pub const LEVELS: usize = 2;

pub struct Game {
    pub player: Player,
    pub levels: Vec<Level>,
}

impl Game {
    pub fn new(p: Player) -> Game {
        let mut v: Vec<Level> = Vec::with_capacity(LEVELS);
        for d in 0..LEVELS {
            v.push(LevelGenerator::generate(d, d == 0, d == LEVELS - 1).render());
        }
        let mut g = Game {
            player: p,
            levels: v,
        };
        let start = {
            g.get_level(0).start
        };
        g.get_mutable_player().get_position().set(0, start.0, start.1);
        g.get_mutable_level(0).discover(&Position::new(0, start.0, start.1));
        g
    }

    pub fn get_player(&self) -> &Player {
        &self.player
    }
    pub fn get_mutable_player(&mut self) -> &mut Player {
        &mut self.player
    }
    pub fn get_level(&self, level: usize) -> &Level { &self.levels[level] }
    pub fn get_mutable_level(&mut self, level: usize) -> &mut Level { &mut self.levels[level] }
    fn can_move(&mut self, dx: i16, dy: i16) -> bool {
        let player_pos = &self.player.get_position();
        let new_x: i16 = player_pos.get_x() as i16 + dx;
        let new_y: i16 = player_pos.get_y() as i16 + dy;
        let level = &mut self.levels[player_pos.get_level()];
        match level.get_element(new_x, new_y) {
            (None, _, _) => { return false; }
            (Some(t), _, _) => {
                match t {
                    StructureElement::Wall => { return false; }
                    _ => {}
                }
            }
        };
        true
    }
    fn next_start(&self) -> (usize, usize, usize) {
        let level = self.player.get_immutable_position().get_level() + 1;
        let (x, y) = self.levels[level].start;
        (level, x, y)
    }
    fn prev_end(&self) -> (usize, usize, usize) {
        let level = self.player.get_immutable_position().get_level() - 1;
        let (x, y) = self.levels[level].end;
        (level, x, y)
    }
    fn next_element(&mut self, dx: i16, dy: i16) -> (Option<StructureElement>, Option<&mut Monster>, Option<&mut Box<(dyn Artifact + 'static)>>) {
        let player_pos = &self.player.get_position();
        let new_x: i16 = player_pos.get_x() as i16 + dx;
        let new_y: i16 = player_pos.get_y() as i16 + dy;
        let player_level = player_pos.get_level();
        let level = &mut self.levels[player_level];
        level.get_element(new_x, new_y)
    }

    pub fn move_player(&mut self, mut dx: i16, mut dy: i16) -> (i16, i16) {
        // verify the player can do the move
        if !self.can_move(dx, dy) {
            return (-dx, -dy);
        }

        let player_pos = &self.player.get_position();
        let mut player_level = player_pos.get_level();
        let (structure, _, _) = self.next_element(dx, dy);

        // use stairs if walked onto one
        match structure {
            Some(e) => {
                match e {
                    StructureElement::StairDown => {
                        (dx, dy) = (0, 0);
                        let (next_level, x, y) = self.next_start();
                        player_level = next_level;
                        self.get_mutable_player().get_position().set(next_level, x, y);
                    }
                    StructureElement::StairUp => {
                        (dx, dy) = (0, 0);
                        let (next_level, x, y) = self.prev_end();
                        player_level = next_level;
                        self.get_mutable_player().get_position().set(next_level, x, y);
                    }
                    _ => {}
                }
            }
            _ => {}
        }

        // update position and discover area on new position
        let (new_x, new_y) = self.get_mutable_player().get_position().change(dx, dy);
        self.get_mutable_level(player_level).discover(&Position::new(player_level, new_x, new_y));
        (-dx, -dy)
    }
    pub fn player_fights_monster(&mut self) -> bool {
        let player_pos = &self.player.get_immutable_position();
        let player_level = player_pos.get_level();
        let level = &mut self.levels[player_level];
        let (_, m, _) = level.get_element(player_pos.get_x() as i16, player_pos.get_y() as i16);
        match m {
            None => {}
            Some(m) => {
                // TODO fight the monster
                self.player.change_life(-1);
                m.decrease_life(1);
                return m.is_dead();
            }
        };
        true
    }
    pub fn player_collects_artifact(&mut self) {
        let player_pos = &self.player.get_immutable_position();
        let player_level = player_pos.get_level();
        let level = &mut self.levels[player_level];
        let (_, _, a) = level.get_element(player_pos.get_x() as i16, player_pos.get_y() as i16);
        match a {
            None => {}
            Some(a) => {
                a.collect(&mut self.player);
            }
        }
    }
    pub fn update_level(&mut self) {
        let player_pos = &self.player.get_immutable_position();
        let player_level = player_pos.get_level();
        let level = &mut self.levels[player_level];
        level.update();
    }
}

#[test]
fn game_has_correct_number_of_levels() {
    let p = Player::new("foo", 42);
    let g = Game::new(p);
    assert_eq!(g.levels.len(), LEVELS);
}

#[test]
fn game_has_player() {
    let p = Player::new("foo", 42);
    let g = Game::new(p);
    assert_eq!(g.get_player().get_name(), "foo");
    assert_eq!(g.get_player().get_immutable_position().get_level(), 0);
}

#[test]
fn game_has_mutable_player() {
    let p = Player::new("foo", 42);
    let mut g = Game::new(p);
    assert_eq!(g.get_player().get_name(), "foo");
    g.get_mutable_player().change_life(-1);
    assert_eq!(g.get_player().get_life(), 41);
}

#[test]
fn game_get_level() {
    let p = Player::new("foo", 42);
    let mut g = Game::new(p);
    g.get_level(0);
    assert_ne!(g.get_level(0).start, (0, 0));
    g.get_mutable_level(0).discover(&Position::new(0, 1, 1));
}
